/*===============================================================
 * Project: libferit
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : 
 * Comment: Call it one time to find a *non-proxy host, then
 *          call it on that host you found to monitor queries
 *
 *-- History -----------------------------------------------
 *
 * 2000-12-25:
 * Mongoose - Created
 *
 ==============================================================*/

#include <time.h>
#include <sys/time.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "Gnutella.h"
#include "Url.h"
#include "utils.h"
#include "Http.h"

// Packet header [0-15][16][17][18][19-22]
#define GUID 0      // Gobally unique identifer
#define FUNC 16     // Gnutella protocol function
#define TTL  17     // Time to live
#define HOPS 18     // Number of hops
#define DATA 19     // Data lenght

#define PING 0x00
#define PONG 0x01
#define PUSH 0x40
#define QURY 0x80
#define HITS 0x81

void print_host(host_t *host)
{
  if (!host)
    return;

  printf("%u.%u.%u.%u:%u [%i, %i]", 
	 host->ip[0], host->ip[1], host->ip[2], host->ip[3],
	 host->port,
	 host->files,
	 host->files_sz);
}

void print_packet(char *type, unsigned char *packet)
{
  int i, len;


  printf("%s[%i][0x%x][%i][%i][%i]\n", type,
	 packet[GUID],
	 packet[FUNC],
	 packet[TTL],
	 packet[HOPS],
	 ((short)packet[DATA+1] << 8) + packet[DATA]);


  switch (packet[FUNC])
  {
  case QURY:
    len = ((short)packet[DATA+1] << 8) + packet[DATA] + DATA+2;
    
    printf("Query::[");

    for (i = DATA+6; i < len; i++)
      printf("%c", packet[i]);

    printf("]\n");
    break;
  }
}


////////////////////////////////////////////////////////////


Gnutella::Gnutella(int timeout) : TcpProtocol(timeout)
{
  _hosts = NULL;
}


Gnutella::~Gnutella()
{
  host_t *tmp;

  while (_hosts)
  {
    printf("Removing host: ");
    print_host(_hosts);
    printf("\n");
    
    tmp = _hosts;
    _hosts = _hosts->next;
    delete tmp;
  }
}



void Gnutella::Ping(int socket)
{
  unsigned char packet[32];


  memset(packet, 0, 32);
  packet[GUID] = 14;
  packet[FUNC] = PING;
  packet[TTL]  = 5;
  packet[HOPS] = 0;

  print_packet("\n\n> ", packet);

  write(socket, packet, 23);
}


void Gnutella::Pong(int socket)
{
  unsigned char packet[40];


  memset(packet, 0, 40);
  packet[GUID] = 14;
  packet[FUNC] = PONG;
  packet[TTL]  = 5;
  packet[HOPS] = 0;
  packet[DATA] = 14;

  //FIXME: Set port, IP, file count, and file size too
  //       For now sending 0.0.0.0:0, 0 files, and 0Kb;
  //       since there is no server  =)

#ifdef DEBUG_PACKETS
  printf(" Port   [%u]\n", ((short)packet[24] << 8) + packet[23]);

  printf(" IP     [%i.%i.%i.%i]\n", 
	 packet[25], packet[26], packet[27], packet[28]);
  
  printf(" Fcount [0x%x 0x%x 0x%x 0x%x]\n", 
	 packet[29], packet[30], packet[31], packet[32]);
  
  printf(" Fsize  [0x%x 0x%x 0x%x 0x%x]\n\n", 
	 packet[33], packet[34], packet[35], packet[36]);
#endif

  print_packet("\n\n> ", packet);  
  write(socket, packet, 36);
}


void Gnutella::Query(int socket, char *query, int min_speed)
{
  unsigned char packet[64];
  int i, len;


  memset(packet, 0, 22);
  packet[GUID] = 14;
  packet[FUNC] = QURY;
  packet[TTL]  = 5;
  packet[HOPS] = 0;

  if (query)
    len = strlen(query);
  else 
  {
    printf("Gnutella::Query> NULL query string\n");
    return;
  }

  // truncate to keep 64 byte limit
  if (len > 39)
    len = 39;

  for (i = 0; i < len; i++)
  {
    packet[i+25] = query[i];
    packet[i+26] = 0;
  }

  // FIXME: Need to allow for > 256  =)
  packet[24] = 0;
  packet[23] = min_speed; 

  //FIXME
  packet[DATA+1] = 0;
  packet[DATA] = len + 5;
  //packet[DATA] = ((short)packet[DATA+1] << 8) + packet[DATA];

  i += 24;

  print_packet("\n\n> ", packet);

  write(socket, packet, i);
}


int Gnutella::Auth(int socket)
{
  char packet[64];
  char buffer[64];
  int i, ii;


  sprintf(packet, "GNUTELLA CONNECT/0.4\n\n");
  printf("> %s\n", packet);
  write(socket, packet, strlen(packet));

  ii = i = 0;

  while (ii < 2 && read(socket, packet, 1))
  {
    if (packet[0] == '\n')
      ii++;
    
    if (i > 63)
    {
      buffer[i] = 0;
      printf("< %s\n", buffer);
      printf("Gnutella::Fetch> Invalid server reply (too long)\n");
      return -2; 
    }
    
    buffer[i++] = packet[0];
  }
  
  buffer[i] = 0;
  printf("< %s\n", buffer);
  
  if (strcmp(buffer, "GNUTELLA OK\n\n") != 0)
  {
    printf("Gnutella::Fetch> Invalid server reply\n");
    return -3;
  }

  return 0;
}


int Gnutella::Download(Url *url, char *filename)
{
   int socket = 0;
   int i;

     
   if (!url->Port())
      url->Port(6346);

   if (!url->Outfile())
      url->Outfile(filename);

   socket = Connect(url->Hostname(), url->Port());

   if (!socket || socket == -1)
   {
     printf("Gnutella::Fetch> Couldn't open socket\n");
     return -1;
   }

   if ((i = Auth(socket)))
     return i;

   // Connected to gnutella server, now send ping packet
   Ping(socket);

   // Now recieve remote host(s) data
   PacketHandler(socket);
 

   // Query testing /////////////////

   printf("\n\nQuery test\n");

   host_t *hosts = _hosts;

   while (hosts)
   {
     char hostname[32];

     sprintf(hostname, "%u.%u.%u.%u", 
	     hosts->ip[0], hosts->ip[1], hosts->ip[2], hosts->ip[3]);

     printf("Connecting to host ");
     print_host(hosts);
     printf(" ...\n\n");

     socket = Connect(hostname, hosts->port);

     if (socket == -1)
       printf("ERROR CONNECTING TO HOST\n\n");
     else
     {
       Query(socket, url->Filename(), 28);
       
       printf("Calling handler\n");
       PacketHandler(socket);
     }

     hosts = hosts->next;
   }

   return 0;
}


/// PRIVATE ///////////////////////////////////////////////

int Gnutella::PacketHandler(int socket)
{
  unsigned char packet_header[23];
  unsigned char *data;
  unsigned char rec[26];
  int i, cc, done;
  host_t *tmp;
  struct timeval tv;
  int timeout = 15;
  fd_set set;


  FD_ZERO(&set);
  FD_SET(socket, &set);
 
  done = 0;

  while (!done)
  {
    i = 0;

    // Set timeout in secs (this is in the loop because linux foos tv)
    tv.tv_sec = timeout;
    tv.tv_usec = 0;  

    cc = select(socket + 1, &set, NULL, NULL, &tv);

    if (!cc)
    {
      printf("Gnutella::PacketHandler> ERROR timeout after %i seconds\n", 
	     timeout);
      return -1;
    }

    while ((cc = read(socket, rec, 1)) && cc > 0)
    {
      packet_header[i] = rec[0];

#ifdef DEBUG_PACKETS
      switch (i)
      {
      case 0:
	printf("\n\n\n GUID   [0x%x ", rec[0]);
	break;
      case 16:
	printf("]\n Func   [0x%x]", rec[0]);
	break;
      case 17:
	printf("\n TTL    [%u]", rec[0]);
	break;
      case 18:
	printf("\n Hops   [%u]\n", rec[0]);
	break;
      case 19:
      case 20:
      case 21:
	break;
      case 22:
	printf(" Data   [%i]\n",
	       ((short)packet_header[20] << 8) + packet_header[19]);
	break;
      default:
	if (i < 22)
	  printf("0x%x ", rec[0]);
      }
#endif
      
      i++;
      
      if (i == 23)
	break;
    }

    if (!(cc > 0))
    {
      done = true;
      continue;
    }

    print_packet("\n< ", packet_header);

    if (packet_header[FUNC] == PING)
      Pong(socket);


    if (packet_header[DATA])
      data = new unsigned char[packet_header[DATA]+4];
    else
      data = NULL;

    i = 0;

    while (data && (cc = read(socket, rec, 1)) && cc > 0)
    {
      data[i] = rec[0];
      
      switch (packet_header[FUNC])
      {
      case PUSH:
	break;
      case PONG:
	switch (i)
        {
#ifdef DEBUG_PACKETS
	case 1:
	  printf(" Port   [%u]\n", ((short)data[1] << 8) + data[0]);
	  break;
	case 5:
	  printf(" IP     [%i.%i.%i.%i]\n", 
		 data[2], data[3], data[4], data[5]);
	  break;
	case 11:
	  printf(" Fcount [0x%x 0x%x 0x%x 0x%x]\n", 
		 data[6], data[7], data[8], data[9]);
	  break;
#endif
	case 13:
#ifdef DEBUG_PACKETS
	  printf(" Fsize  [0x%x 0x%x 0x%x 0x%x]\n\n", 
		 data[10], data[11], data[12], data[13]);
#endif	  

	  if (!_hosts)
	  {
	    _hosts = new host_t;
	    _hosts->next = NULL;
	  }
	  else  // prepend new host
 	  {
	    tmp = _hosts;
	    _hosts = new host_t;
	    _hosts->next = tmp;	   
	  }
	  
	  _hosts->ip[0] = data[2];
	  _hosts->ip[1] = data[3];
	  _hosts->ip[2] = data[4];
	  _hosts->ip[3] = data[5];
	  _hosts->port = ((short)data[1] << 8) + data[0];
	  _hosts->files = ((int)data[6] << 24) + ((int)data[7] << 16) + 
	                  ((int)data[8] << 8) + data[9];
	  _hosts->files_sz = ((int)data[10] << 24) + ((int)data[11] << 16) + 
	                     ((int)data[12] << 8) + data[13]; 

	  printf("New host ");
	  print_host(_hosts);
	  printf("\n");
	  break;
	}
	break;
      case QURY:
	if (i == 0)
	  printf("\n QUERY Recieved \'");
	
	printf("%c", rec[0]);
	
	if (i == packet_header[DATA]-1)
	  printf("\'\n");
	break;  
      default:
	if (!i)
	  fprintf(stderr, "UNKNOWN FUNCTION 0x%x\n\'", packet_header[FUNC]);

	fprintf(stderr, "%c", rec[0]);

	if (i == packet_header[DATA]-1)
	  fprintf(stderr, "\'\n");
      }
      
      i++;
      
      if (i >= packet_header[DATA])
	break;
    }
    
    if (data)
      delete [] data;
    
    if (!(cc > 0))
      done = true;
  }
  
  close(socket);

  return 0;
}
