/*===============================================================
 * Project: libferit ( base netcode from Snarf 2.0 )
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Http
 * Comment: Http module.
 *
 *-- History -----------------------------------------------
 *
 * 2000-02-14:
 * Mongoose - return server error code as int
 * 
 * 2000-01-29:
 * Mongoose - Created
 *
 ==============================================================*/

#define DEBUG_HTTP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <pwd.h>

#include "Http.h"

Http::Http(int timeout) : TcpProtocol(timeout)
{
  _redirect_count = 0;
}


Http::~Http()
{
}


int Http::Download(Url *url, char *filename)
{
   Url *proxy;
   Url *redir_u;
   char *header		= NULL;
   char *lc_header	= NULL;
   char *len_string 	= NULL;
   char *new_location	= NULL;
   char buf[BUFSIZE];
   int sock 		= 0;
   ssize_t bytes_read	= 0;
   int retval		= 0;
   int i;
   char s_error[4];

   
   // FIXME: Replace this way of doing redirects
   /* Make sure we haven't recursed too much */
   if (_redirect_count > REDIRECT_MAX) 
   {
     printf("Http::Download> ERROR: Max redirection count exceeded.\n");
     _redirect_count = 0;
     return -1;
   }

   if (filename)
   {
      url->Outfile(filename);
   }

   /* Make sure everything's initialized to something useful */
   if (!url->Hostname()) 
   {
      printf("ERROR: no host specified");
      return -2;
   }

   //FIXME: Handle this a little better
   if (!url->Outfile()) 
   {
      if (url->Filename()) 
         url->Outfile(url->Filename());
      else
         url->Outfile("index.html");
   }

   /* Send the request to either the proxy or the remote host */
   if (!(url->ProxyType() == NO_PROXY)) 
   {
      proxy = url->Proxy();

      if (!proxy || !proxy->Hostname()) 
      {
         printf("Http::Download> ERROR: Bad proxy url\n");
         return -3;
      }

      /* Prompt for proxy password if not specified */
      if (proxy->Username() && !proxy->Password()) 
      {
         char *prompt = NULL;
         prompt = strconcat("Password for proxy ",
                            proxy->Username(), "@",
                            proxy->Hostname(), ": ", NULL);
         proxy->Password(getpass(prompt));
         delete prompt;
      }

      if (!(sock = Connect(proxy->Hostname(), proxy->Port())))
         return -4;
   } 
   else /* No proxy */ 
   {
      if (!(sock = Connect(url->Hostname(), url->Port())))
         return -5;
   }
    
   if (Request(sock, url))
     return -51;

   /* Check to see if it returned a HTTP 1.x response */
   memset(buf, '\0', 5);

   bytes_read = read(sock, buf, 8);

   if (url->Options() & OPT_VERBOSE) 
     fprintf(stderr, "\nHost replied:\n\n");

   if (bytes_read == 0) 
   {
      close(sock);
      return -6;
   }

   if (!(buf[0] == 'H' && buf[1] == 'T' && buf[2] == 'T' && buf[3] == 'P')) 
   {
      if ((url->Options() & OPT_RESUME) && url->OutfileOFF()) 
      {
	printf("WARNING: server does not support resume.");
        retval = -7;
        close(sock);
        return retval;
      }
        
      //FIXME: ah, kind of made it hard to use this now  =)        
      //write(fileno(out), buf, bytes_read);
   } 
   else 
   {
      /* Skip the header */
      buf[bytes_read] = '\0';
      header = strdup(buf);
      header = GetHeader(sock);
      header = strconcat(buf, header, NULL);
                
      lc_header = strdup(header);
      lc_header = string_lowercase(lc_header);

      if (url->Options() & OPT_VERBOSE) 
         fwrite(header, 1, strlen(header), stderr);
                
      /* Check for redirects */
      new_location = GetHeaderValue("location", lc_header);

      if (header[9] == '3' && new_location) 
      {
         redir_u = new Url(new_location);

         if (!redir_u)
	 {
	    printf("ERROR: couldn't make Url for: %s\n", new_location);
            close(sock);
            return -8;
	 }                
        
         /* Make sure we still send user/password along */
         if (url->Username())
	 {
            redir_u->Username(url->Username());
            redir_u->Password(url->Password());
	 }

         _redirect_count++;
         retval = Download(redir_u, filename);
         close(sock);
         return retval;
      }

      if (header[9] == '4' || header[9] == '5') 
      {
         for (i = 0; header[i] && header[i] != '\n'; i++)
            ;

         header[i] = '\0';
         printf("ERROR: HTTP error from server: %s\n", header);

	 close(sock);

         /* Mongoose: Buffer error code as string error */
         s_error[0] = header[9];
         s_error[1] = header[10];
         s_error[2] = header[11];
         s_error[3] = 0;

         /* Mongoose: Translate string error to int error */
         retval = atoi(s_error);

         return retval;
      }
                        
      len_string = GetHeaderValue("content-length", lc_header);

      if (len_string)
         url->OutfileSZ((off_t )atoi(len_string));

      if (GetHeaderValue("content-range", lc_header))
         url->OutfileSZ(url->OutfileSZ() + url->OutfileOFF());

      if ((!url->OutfileSZ()) && 
          (url->Options() & OPT_RESUME) && !(url->Options() & OPT_NORESUME)
          && url->OutfileOFF()) 
      {
         printf("WARNING: unable to determine remote file size.\n");
	 close(sock);
         return -9;
      }
   }

   return SaveStream(sock, url);
}
                        

char *Http::GetHeaderValue(char *header, const char *full_header)
{
   char *value		= NULL;
   char *retval        	= NULL;
   char *actual_header 	= NULL;
   int i;


   if (header == NULL || full_header == NULL)
      return NULL;

   actual_header = strconcat("\n", header, NULL);

   if ((value = strstr(full_header, actual_header))) 
   {
      /* Skip the preceding newline and trailing colon and space */
      value += strlen(actual_header) + 2;
                
      /* Find the end of the line */
      for(i = 0; value[i] && value[i] != '\r' && value[i] != '\n'; i++)
         ;

      retval = new char[i + 1];
      memcpy(retval, value, i);

      retval[i + 1] = '\0';
      return retval;
   } 
   else 
      return NULL;
}

//FIXME: I agree with "This whole function is pathetic..."
//FIXME: I also agree with "...please rewrite it for me."  ;)
char *Http::GetHeader(int fd)
{
   char *header = NULL;
   char buffer[BUFSIZE]; 	
   int bytes_read = 0;
   int total_read = 0;

   header = new char[2];
   header[0] = header[1] = 0;

   buffer[0] = buffer[1] = buffer[2] = '\0';

   while((bytes_read = read(fd, buffer, 1))) 
   {
      total_read += bytes_read;

      header = strconcat(header, buffer, NULL);

      if (total_read > 1) 
      {
         if (strcmp(header + (total_read - 2), "\n\n") == 0)
            return header;
      }

      if (total_read > 3) 
      {
         if (strcmp(header + (total_read - 4), "\r\n\r\n") == 0)
            return header;
      }
   }

   return header;
}


int Http::Request(int socket, Url *url)
{
  off_t file_size;
  char *request = NULL;
  char *auth = NULL;


  // Mongoose: File request
  if (url->Proxy())
  {
    BufferMessage(&request, "GET %s HTTP/1.0\r\n", url->FullUrl(), NULL);
    BufferMessage(&request, "Host: %s\r\n", url->Hostname(), NULL);
  }      
  else
  {
    if (url->Filename())
    {
      BufferMessage(&request, "GET /%s%s HTTP/1.0\r\n", 
		    url->Path(), url->Filename(), NULL);
    }
    else
    {
      BufferMessage(&request, "GET / HTTP/1.0\r\n", NULL);
    }
 
    BufferMessage(&request, "Host: %s\r\n", url->Hostname(), NULL);
  }

  // Mongoose: Auth
  if (url->Username() && url->Password()) 
  {
    auth = strconcat(url->Username(), ":", url->Password(), NULL);
    auth = base64(auth, strlen(auth));
    BufferMessage(&request, "Authorization: Basic %s\r\n", auth, NULL);
  }

  // Mongoose: Proxy Auth
  if (url->Proxy() && url->Proxy()->Username() && url->Proxy()->Password()) 
  {
    auth = strconcat(url->Proxy()->Username(), ":", 
		     url->Proxy()->Password(), NULL);
    auth = base64(auth, strlen(auth));
    BufferMessage(&request, "Proxy-Authorization: Basic %s\r\n", auth, NULL);
  }

   // Mongoose: Resume
   if ((url->Options() & OPT_RESUME) && url->Outfile())
   {
     file_size = get_file_size(url->Outfile());

     if (file_size > 0)
     {
       char num[64];

       sprintf(num, "%ld", (long int)file_size);
       BufferMessage(&request, "Range: bytes=%s-\r\n", num, NULL);
     }
   }

   // Mongoose: Send User-Agent string
   if (getenv("FAKE_HTTP_USER_AGENT"))
      BufferMessage(&request, "User-Agent: %s\r\n", 
		    getenv("FAKE_HTTP_USER_AGENT"), NULL);
   else if (url->Options() & OPT_BE_MOZILLA) 
      BufferMessage(&request, "User-Agent: %s\r\n", MOZILLA_USER_AGENT, NULL);
   else if (url->Options() & OPT_BE_MSIE)
      BufferMessage(&request, "User-Agent: %s\r\n", MSIE_USER_AGENT, NULL);
   else if (url->Options() & OPT_BE_SNARF)
      BufferMessage(&request, "User-Agent: %s\r\n", SNARF_USER_AGENT, NULL);
   else 
      BufferMessage(&request, "User-Agent: %s\r\n", USER_AGENT, NULL);

   // Mongoose: Send Referer string 
   if (getenv("FAKE_HTTP_REFERER"))
      BufferMessage(&request, "Referer: %s\r\n", 
		    getenv("FAKE_HTTP_REFERER"), NULL);
   
   BufferMessage(&request, "\r\n", NULL);

   if (url->Options() & OPT_VERBOSE) 
     printf("Sending request:\n%s\n", request);

   // Mongoose: Write to socket    
   if (write(socket, request, strlen(request)) == -1)
   {
     perror("Http::Request> ");
     return -1;
   }
   else if (url->Options() & OPT_VERBOSE)
     printf("\nSucessful Request\n");

   return 0;
}


// Mongoose: Arg list must end with NULL arg
int Http::BufferMessage(char **request, const char *string, ...)
{
  unsigned int len;
  char *msg = NULL;
  char *swap = NULL;
  char *vs = NULL;
  va_list args;

  
  if (!string)
  {
    printf("Http::BufferMessage> ERROR Bad message\n");
    return -2;
  }

  len = strlen(string);

  // Mongoose: Find length of generated substrings
  va_start(args, string);
  vs = va_arg(args, char *);

  while (vs)
  {
    len += strlen(vs);
    vs = va_arg(args, char *);
  }
   
  va_end(args);

  msg = new char[len+1];
  msg[0] = 0;

  // Mongoose: Buffer message into a single string 
  va_start(args, string);
  vsprintf(msg, string, args);
  va_end (args);

  if (*request)
  {
    swap = new char[strlen(msg) + strlen(*request) + 1];
    swap[0] = 0;

    strcpy(swap, *request);
    strcat(swap, msg);

    delete [] *request;
    delete [] msg;

    *request = swap;
  }
  else
    *request = msg;


  return 0;
}


// Mongoose: Arg list must end with NULL arg
int Http::Message(int socket, const char *s, ...)
{
  unsigned int len;
  char *msg = NULL;
  char *vs = NULL;
  va_list args;

  
  if (socket < 0)
  {
    printf("Http::Message> ERROR Bad socket\n");
    return -1;
  }

  if (!s)
  {
    printf("Http::Message> ERROR Bad message\n");
    return -2;
  }

  len = strlen(s);

  // Mongoose: Find length of generated substrings
  va_start(args, s);
  vs = va_arg(args, char *);

  while (vs)
  {
    len += strlen(vs);
    vs = va_arg(args, char *);
  }
   
  va_end(args);

  printf("*** len = %u\n", len);

  msg = new char[len+1];

  if (!msg)
  {
    perror("Http::Message> ");
    return -3;
  }

  msg[0] = 0;

  // Mongoose: Buffer message into a single string 
  va_start(args, s);
  vsprintf(msg, s, args);
  va_end (args);

  //printf("*** msg = '%s'\n", msg);
  
  if (write(socket, msg, strlen(msg)) == -1)
  {
    perror("Http::Message> ");
    return -4;
  }

  delete [] msg;
  printf("*** cleared msg\n");

#ifdef DEBUG
  printf("Http::Message> Wrote '%s' %i bytes\n", msg, strlen(msg));
#endif

  return 0;
}



