/*===============================================================
 * Project: libferit ( base netcode from Snarf 2.0 )
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : Progress
 * Comment: Progress module.
 *
 *-- History -----------------------------------------------
 *
 * 2000 01 30:
 * Mongoose - Created
 *
 ==============================================================*/

#include "Progress.h"
#include "utils.h"
#include <sys/ioctl.h>
#include <sys/time.h>
#include <ctype.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

char OUTPUT_BUFFER[BUFSIZE];


void repchar(FILE *fp, char ch, int count)
{
   while(count--)
      fputc(ch, fp);
}

Progress::Progress()
{
   _tty	        = false;
   _length 	= 0;
   _current 	= 0;
   _offset	= 0;
   _overflow	= 0;
   _max_hashes  = 0;
   _cur_hashes  = 0;
   _url         = NULL;
   _outfile     = NULL;
   _start_time  = Time();
}

// FIXME: Segfaults on last delete []
Progress::~Progress()
{
   double elapsed = 0;
   double kbytes;


   if (!_tty)
      return;

   elapsed = Time() - _start_time;

   fprintf(stderr, "\n");

   if (elapsed) 
   {
      kbytes = ((float )(_current - _offset) / elapsed) / 1024.0;

      fprintf(stderr, "%ld bytes transferred in %.2f sec (%.2fk/sec)\n",
              _current - _offset, elapsed, kbytes);
   } 
   else 
   {
      fprintf(stderr, "%ld bytes transferred in less than a second\n", 
              _current);
   }

   if (_outfile)
     delete [] _outfile;

   fflush(stderr);
}

int Progress::Winsize()
{
#ifdef _SYS_IOCTL_H
   int width;
   struct winsize ws;


   if (ioctl(STDERR_FILENO, TIOCGWINSZ, &ws) == -1 || ws.ws_col == 0 )
      width = 79;
   else
      width = ws.ws_col - 1;

   return width;
#else
   return 79;
#endif
}

double Progress::Time()
{
   struct timeval tv;
   double retval;
     

   gettimeofday(&tv, NULL);

   retval = tv.tv_sec + (tv.tv_usec / 1000000.00);

   return retval;
}

int Progress::Init(Url *url, long int len)
{
   char *filename 	= NULL;
   int win_width	= 0;
   int total_units	= 0;


   _url = url;

   if (_url->Options() & OPT_PROGRESS)
      _tty = true;
   else 
   {
      _tty = false;
      return 1;
   }
   
   if (_url->Outfile()) 
      _outfile = mon_strdup(_url->Outfile());

   win_width = Winsize();

   if (win_width > 30)
      total_units = win_width - (30 + 9 + 16);
   else 
   {
      /* The window is pathetically narrow; bail */
      total_units = 20;
   }

   _length = len;

   /* Buffering stderr: no flicker! yay! */
   setbuf(stderr, (char *)&OUTPUT_BUFFER);

   fprintf(stderr, "%s (", _url->FullUrl());

   if (total_units && len) 
   {
      _length = len;
      _max_hashes = total_units;
      fprintf(stderr, "%dK", (int )(len / 1024));
   } 
   else
      fprintf(stderr, "unknown size");

   fprintf(stderr, ")\n");

   _current = 0;

   filename = mon_strdup(_outfile);
   
   if (strlen(filename) > 24)
      filename[24] = '\0';

   fprintf(stderr, "%-25s[", filename);

   if (_length)
      repchar(stderr, ' ', _max_hashes);
   else
      fputc('+', stderr);

   fprintf(stderr, "] %7dK", (int )_current);

   if (filename)
      delete [] filename;

   fflush(stderr);

   return 1;
}


void Progress::Update(long int increment)
{
   unsigned int units;
   //char *anim = "-\\|/";
   char *anim = "-+*#*+";


   if (!_tty)
      return;

   _current += increment;

   if (strlen(_outfile) > 24) 
      _outfile[24] = '\0';

   fprintf(stderr, "\r");
   fprintf(stderr, "%-25s [", _outfile);
   

   if (_length ) 
   {
      float percent_done = (float )_current / _length;
      double elapsed;
      float rate;

      // added (unsigned int) cast
      units = (unsigned int)(percent_done * _max_hashes);

      if (units)
         repchar(stderr, '#', units);

      repchar(stderr, ' ', _max_hashes - units);
      fprintf(stderr, "] ");
      fprintf(stderr, "%7dK", (int )(_current / 1024));
       
      elapsed = Time() - _start_time;

      if (elapsed) 
         rate = ((_current - _offset) / elapsed) / 1024;
      else
         rate = 0;

      /* The first few runs give extra-high values, so skip them */
      if (rate > 999999)
         rate = 0;

      fprintf(stderr, " | %7.2fK/s", rate);      
   } 
   else 
   {
      /* Length is unknown, so just draw a little spinny thingy */
      fprintf(stderr, "%c]", anim[_frame++ % 4]);
      fprintf(stderr, " %7dK", (int )(_current / 1024));
   }

   fflush(stderr);
   return;
}

void Progress::Offset(long int n)
{
   _offset = n;
}
