/*===============================================================
 * Project: libferit
 * Author : Terry 'Mongoose' Hendrix
 * Website: http://www.westga.edu/~stu7440/
 * Email  : stu7440@westga.edu
 * Object : TcpProtocol
 * Comment: TCP wrapper class
 *
 *-- History -----------------------------------------------
 *
 * 2001-01-13:
 * Mongoose - Removing more snarf based code
 *
 * 2000-02-14:
 * Mongoose - return 0 if have full file too
 *
 * 2000-01-29:
 * Mongoose - Created, based on snarf http client
 *
 ==============================================================*/

#include <time.h>
#include <sys/time.h>
#include "TcpProtocol.h"
#include "Progress.h"

TcpProtocol::TcpProtocol(int timeout)
{
  if (!timeout)
    timeout = 5;

  _timeout = timeout;
  _close = true;
}


TcpProtocol::~TcpProtocol()
{
}


void TcpProtocol::DisableClose()
{
  _close = false;
}


int TcpProtocol::SaveStream(int socket, Url *url)
{
  FILE *f = NULL;
  int i, n, cc;
  struct timeval tv;
  fd_set set;


  if (socket == -1)
  {
     printf("TcpProtocol::SaveStream> Bad socket passed!\n");
     return -1;
  }

  FD_ZERO(&set);
  FD_SET(socket, &set);
 
  // Timeout tv set to N secs (remember linux foos tv after select call)
  tv.tv_sec = _timeout;
  tv.tv_usec = 0;  
  
  cc = select(socket + 1, &set, NULL, NULL, &tv);
  
  if (!cc)
  {
    fprintf(stderr, "TCP::SaveStream> ERROR no data in %i seconds\n", _timeout);
    return -1;
  }


  if (!url->Outfile())
  {
    char filename[32];


    printf("TCP::SaveStream> WARNING: NULL output filename.\n");
    
    for (i = 0; i < 9999; i++)
    {
      if (i < 10)
	sprintf(filename, "ferit-000%i", i);
      else if (i < 100)
	sprintf(filename, "ferit-00%i", i);
      else if (i < 1000)
	sprintf(filename, "ferit-0%i", i);
      else
	sprintf(filename, "ferit-%i", i);
      
      f = fopen(filename, "r");

      if (!f)
      {
	url->Outfile(filename);
	break;
      }
      else
	fclose(f);
    }

    if (i >= 9999)
      printf("TCP::SaveStream> ERROR: Couldn't generate filename not in use\n");
    else
      printf("TCP::SaveStream> Generated filename %s.\n", url->Outfile());
  }

  // Append if resuming, else start a new file or truncate
  if (url->Options() & OPT_RESUME && !(url->Options() & OPT_NORESUME)) 
    f = fopen(url->Outfile(), "a");
  else
    f = fopen(url->Outfile(), "w");

  if (!f) 
  {
    perror("TCP::SaveStream> ERROR");

    if (_close)
      close(socket);

    return -2;
  }
  
  n = SaveData(url, socket, f);
  
  //FIXME: Handle errors, that can be, later perhaps
  
  return n;
}


int TcpProtocol::PrintStream(int socket, Url *url)
{
  int cc;
  struct timeval tv;
  fd_set set;
  char c[2];


  if (socket == -1)
  {
     printf("TcpProtocol::PrintStream> Bad scoket passed!\n");
     return -1;
  }

  FD_ZERO(&set);
  FD_SET(socket, &set);
 
  // Timeout tv set to N secs (remember linux foos tv after select call)
  tv.tv_sec = _timeout;
  tv.tv_usec = 0;  
  
  cc = select(socket + 1, &set, NULL, NULL, &tv);
  
  if (!cc)
  {
    fprintf(stderr, "TCP::SavePrint> ERROR no data in %i seconds\n", _timeout);
    return -1;
  }

  while (read(socket, c, 1) > 0) 
  {
    printf("%c", c[0]);
  }

  printf("\n\n");

  if (_close)
    close(socket);   
  
  return 0;
}


void TcpProtocol::Timeout(int timeout)
{
  _timeout = timeout;
}


int TcpProtocol::Connect(char *remote_host, int port) 
{
   struct hostent *host;
   struct sockaddr_in sa;
   int sock_fd;


   if ((host = (struct hostent *)gethostbyname(remote_host)) == NULL) 
   {
      herror(remote_host);
      return -1;
   }

   /* Get the socket */
   if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
   {
      perror("socket");
      return -1;
   }

   /* Connect the socket, filling in the important stuff */
   sa.sin_family = AF_INET;
   sa.sin_port = htons(port);
   memcpy(&sa.sin_addr, host->h_addr, host->h_length);
  
   if (connect(sock_fd, (struct sockaddr *)&sa, sizeof(sa)) < 0)
   {
      perror(remote_host);
      return -1;
   }

   return sock_fd;
}


/// Private //////////////////////////////////////


int TcpProtocol::SaveData(Url *url, int sock, FILE *out)
{
  off_t rfile_sz, afile_sz, file_off;
  int out_fd = fileno(out);
  int bytes_read = 0;
  ssize_t written = 0;
  char buf[BUFSIZE];
  Progress *p = NULL;


  rfile_sz = url->OutfileSZ();
  afile_sz = get_file_size(url->Outfile());
  file_off = url->OutfileOFF();

  if (url->Options() & OPT_VERBOSE)
    printf("TcpProtocol::SaveData> Offset=%li ReportedSZ=%li ActualSZ=%li\n", 
	   file_off, rfile_sz, afile_sz);

  if (afile_sz && afile_sz != rfile_sz)
  {
    printf("TcpProtocol::SaveData> WARNING: File size ACTUAL=%li REPORTED=%li\n",
	   afile_sz, rfile_sz);
  }

  if (!(url->Options() & OPT_NORESUME)) 
  {  
    if (afile_sz && (afile_sz >= rfile_sz)) 
    {
      printf("TcpProtocol::SaveData> WARNING: Already have all of `%s'.\n", 
	     url->Outfile());
         
      if (_close)
	close(sock);

      return 0;
    }
  }

  if (!(url->Options() & OPT_QUIET))
  {
    p = new Progress();
    p->Init(url, rfile_sz);
        
    if (!(url->Options() & OPT_NORESUME)) 
    {
      p->Update(file_off);
      p->Offset(file_off);
    }
  }

  while ((bytes_read = read(sock, buf, BUFSIZE)) ) 
  {
    if (p)
      p->Update(bytes_read);     
           
    written = write(out_fd, buf, bytes_read);
                 
    if (written == -1) 
    {
      perror("TcpProtocol::SaveData> ERROR write():");

      if (_close)
	close(sock);

      return -1;
    }
  }

  if (_close)
    close(sock);

  if (p)
    delete p;

  return 0;
}

